/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing author: Naveen Michaud-Agrawal (Johns Hopkins U)
------------------------------------------------------------------------- */

#include "mpi.h"
#include "string.h"
#include "compute_group_groups_atom.h"
#include "atom.h"
#include "update.h"
#include "comm.h"
#include "force.h"
#include "pair.h"
#include "memory.h"
#include "neighbor.h"
#include "neigh_request.h"
#include "neigh_list.h"
#include "group.h"
#include "domain.h"
#include "error.h"

using namespace LAMMPS_NS;

/* ---------------------------------------------------------------------- */

ComputeGroupGroupsAtom::ComputeGroupGroupsAtom(LAMMPS *lmp, int narg, char **arg) : 
  Compute(lmp, narg, arg)
{
  if (narg < 4) error->all(FLERR,"Illegal compute group/groups/atom command");

  //My modifications
  nJgroups = narg-3;
  oGroups = new char*[nJgroups];
  jgroups = new int[nJgroups];
  jgroupbits = new int[nJgroups];

  peratom_flag=1;
  size_peratom_cols=4*(nJgroups+1);
  array_flag=1;
  vector_flag=1;
  
  size_vector = size_peratom_cols;
  vector = new double[size_vector];
  comm_reverse = size_vector;

  //End My Modifications
  for (int i=0; i<nJgroups; ++i) {
	  int n = strlen(arg[3+i]) + 1;
	  oGroups[i] = new char[n];
	  strcpy(oGroups[i],arg[3+i]);

	  jgroups[i] = group->find(oGroups[i]);
	  if (jgroups[i] == -1) 
		error->all(FLERR,"Compute group/groups/atom group ID does not exist");
	  jgroupbits[i] = group->bitmask[jgroups[i]];
	 //if (screen) fprintf(screen,"%s in group/groups/atom compute\n",oGroups[i]);
  }
  
  jGrpCheck = 0;
  for (int k = 0; k < nJgroups; ++k)
	  jGrpCheck|=jgroupbits[k];
  nmax=0;
  array_atom=NULL;
  array=NULL;
  
  
}

/* ---------------------------------------------------------------------- */

ComputeGroupGroupsAtom::~ComputeGroupGroupsAtom()
{
  delete [] array;
  for (int i=0; i<nJgroups; ++i) {
	  delete [] oGroups[i];
  }
  delete [] oGroups;
  delete [] jgroups;
  delete [] jgroupbits;
  delete [] vector;
}

/* ---------------------------------------------------------------------- */

void ComputeGroupGroupsAtom::init()
{
  if (force->pair == NULL)
    error->all(FLERR,"No pair style defined for compute group/groups/atom");

  // if non-hybrid, then error if single_enable = 0
  // if hybrid, let hybrid determine if sub-style sets single_enable = 0

  if (force->pair_match("hybrid",0) == NULL && force->pair->single_enable == 0)
    error->all(FLERR,"Pair style does not support compute atom/groups");
  if (!(force->newton_pair))
    error->all(FLERR,"Newton pairing must be turned on for compute group/groups/atom");
  
  pair = force->pair;
  cutsq = force->pair->cutsq;

  // recheck that group 2 has not been deleted

  for (int i = 0; i<nJgroups; ++i) {
	  jgroups[i] = group->find(oGroups[i]);
	  if (jgroups[i] == -1) 
		error->all(FLERR,"Compute group/groups/atom group ID does not exist");
	  jgroupbits[i] = group->bitmask[jgroups[i]];
  }
  jGrpCheck = 0;
  for (int k = 0; k < nJgroups; ++k)
	  jGrpCheck|=jgroupbits[k];

  // need an occasional half neighbor list

  int irequest = neighbor->request((void *) this);
  neighbor->requests[irequest]->pair = 0;
  neighbor->requests[irequest]->compute = 1;
  neighbor->requests[irequest]->occasional = 1;
}

/* ---------------------------------------------------------------------- */

void ComputeGroupGroupsAtom::init_list(int id, NeighList *ptr)
{
  list = ptr;
}

/* ---------------------------------------------------------------------- */

void ComputeGroupGroupsAtom::compute_peratom()
{
  if (atom->nmax > nmax) {
    nmax = atom->nmax; 
    memory->destroy(array);
    memory->create(array,nmax,size_peratom_cols,"group/groups/atom:peratom");
    array_atom = array;
  }
  invoked_peratom = update->ntimestep;
  invoked_vector = update->ntimestep;

  interact();
}

void ComputeGroupGroupsAtom::compute_vector()
{
  if (atom->nmax > nmax) {
    nmax = atom->nmax; 
    memory->destroy(array);
    memory->create(array,nmax,size_peratom_cols,"group/groups/atom:peratom");
    array_atom = array;
  }
  invoked_peratom = update->ntimestep;
  invoked_vector = update->ntimestep;

  interact();
}

/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
void ComputeGroupGroupsAtom::interact()
{
  //outside needed data
  double **x = atom->x;
  int *type = atom->type;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;
  double *special_coul = force->special_coul;
  double *special_lj = force->special_lj;
  int newton_pair = force->newton_pair;
  neighbor->build_one(list->index);

  //list data
  int inum = list->inum;
  int *ilist = list->ilist;
  int *numneigh = list->numneigh;
  int **firstneigh = list->firstneigh;

  //temperorary constants
  int *jlist;
  int i, j, k, m, k4, atom1, atom2, a[2], t[2], jnum;
  double factor_lj, factor_coul, rsq, epair, fpair, del[3];

  //storage
  double *one = new double[size_vector];
  double *all = new double[size_vector];
  for (i = 0; i<size_vector; ++i)
	  one[i]=0;
  for (k = 0; k < nmax; ++k) {
	  for (i = 0; i < size_vector; ++i) 
		  array[k][i] = 0;
  }
  int allgroups = groupbit | jGrpCheck;
  for (i = 0; i < inum; ++i) {
	  atom1 = ilist[i];
    if (!(mask[atom1]&allgroups)) continue;
    jlist = firstneigh[atom1];
    jnum = numneigh[atom1];
    for (j = 0; j < jnum; ++j) {
        atom2 = jlist[j];
        factor_lj = special_lj[sbmask(atom2)];
        factor_coul = special_lj[sbmask(atom2)];
        atom2 &= NEIGHMASK;         //strip neighbor flags off of atom2
        if (!(mask[atom2]&allgroups)) continue; //at this point, both atoms could be in main group or jgroups, make sure both are represetned
        if (mask[atom1]&groupbit && mask[atom2]&jGrpCheck) { 
            a[0] = atom1;
            a[1] = atom2;
        } else if (mask[atom2]&groupbit && mask[atom1]&jGrpCheck) { 
            a[0] = atom2;
            a[1] = atom1;
        } else
            continue; //neither atom is in main group
        t[0] = type[a[0]];
        t[1] = type[a[1]];
        for (m = 0; m < 3; ++m)
            del[m] = x[a[0]][m] - x[a[1]][m];
        domain->minimum_image(del);
        rsq = del[0]*del[0] + del[1]*del[1] + del[2]*del[2];
        epair = pair->single(a[0],a[1],t[0],t[1],rsq,factor_coul,factor_lj,fpair)*0.5;
        for (m = 0; m < 3; ++m)
            del[m] *= fpair;
        for (k = 0; k < nJgroups; ++k) {
            k4 = 4*k;
            if (mask[a[1]] & jgroupbits[k]) {
                if (jgroupbits[k] & groupbit) {
                    array[a[0]][k4] += epair;
                    array[a[1]][k4] += epair;
                    one[k4]         += 2.0*epair;
                    for (m = 0; m < 3; ++m) {
                        array[a[0]][k4+m+1] += del[m];
                        array[a[1]][k4+m+1] -= del[m];
                    }
                } else {
                    array[a[0]][k4] += epair;
                    one[k4]         += epair;
                    for (m = 0; m < 3; ++m) {
                        array[a[0]][k4+m+1] += del[m];
                        one[k4+m+1]         += del[m];
                    }
                }
            }
        }
        k4 = nJgroups*4;
        if (jGrpCheck & groupbit && mask[a[1]] & groupbit) {
            array[a[0]][k4] += epair;
            array[a[1]][k4] += epair;
            one[k4]         += 2.0*epair;
            for (m = 0; m < 3; ++m) {
                array[a[0]][k4+m+1] += del[m];
                array[a[1]][k4+m+1] -= del[m];
            }
        } else {
            array[a[0]][k4] += epair;
            one[k4]         += epair;
            for (m = 0; m < 3; ++m) {
                array[a[0]][k4+m+1] += del[m];
                one[k4+m+1]         += del[m];
            }
        }
    }
  }

  comm->reverse_comm_compute(this);
  //need to aggregate per atom data per node...
  MPI_Allreduce(one,all,size_vector,MPI_DOUBLE,MPI_SUM,world);
  for (i = 0; i < size_vector; ++i)
	  vector[i] = all[i];
  delete [] one;
  delete [] all;
}

/* ---------------------------------------------------------------------- */

int ComputeGroupGroupsAtom::pack_reverse_comm(int n, int first, double *buf)
{
  int i,m,k,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
	  for (k = 0; k < size_vector; ++k)
		  buf[m++] = array[i][k];
  }
  return size_vector;
}

/* ---------------------------------------------------------------------- */

void ComputeGroupGroupsAtom::unpack_reverse_comm(int n, int *llist, double *buf)
{
  int i,j,k,m;

  m = 0;
  for (i = 0; i < n; i++) {
    j = llist[i];
	for (k = 0; k < size_vector; ++k) 
		array[j][k] += buf[m++];
  }
}

double ComputeGroupGroupsAtom::memory_usage()
{
  double bytes = nmax * size_vector * sizeof(double) + size_vector * sizeof(double);  //Per-atom data plus global vector
  return bytes;
}
